/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2013-2014 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of the GNU General
 * Public License Version 2 only ("GPL"). You may not use this file except
 * in compliance with the License.  You can obtain a copy of the License at
 * https://avatar.java.net/license.html or legal/LICENSE.txt.  See the
 * License for the specific language governing permissions and limitations
 * under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 */

/**
 * @module org/glassfish/avatar
 */

'use strict';

var common = require("org/glassfish/avatar/common");
var internal = require("org/glassfish/avatar/internal");
var dataproviders = require("org/glassfish/avatar/dataproviders");
var jms = require("org/glassfish/avatar/jms");
var url = require('url');
var registry = internal.registry;
var log = common.log;
var logFine = common.logFine;
var willLogFine = common.willLogFine;

/**
 * @name RestService
 * @class
 * @classdesc A RestService encapsulates the logic for getting and manipulating a specific resource or collection of resources.
 * It's exposed to the browser as a URL or URL pattern and supports some or all of the HTTP methods GET, PUT, POST, and DELETE.
 * A RestService can be implemented imperatively by overriding one or several of the methods onGet, onPut, onPost, and onDelete,
 * or declaratively by binding it to data providers, or a combination of the two.
 * @see {@link module:org/glassfish/avatar.registerRestService registerRestService}
 *
 */
var RestService = function() {

        /**
         * The service method is called for each request that arrives regardless of method. The default implementation
         *  deserializes the content of the request and dispatches the request to the appropriate onXXX method.
         *  In most scenarios there is no need to override this method.
         * @memberof RestService
         */
        this.service = function(request, response) {
            var doMethod = "on" + request.method;
            var data = null;
            if (request.content != null) {
                if (request.contentType.indexOf('application/x-www-form-urlencoded') >= 0) {
                    data = url.parse('?' + request.content, true).query;
                } else if (request.contentType.indexOf('application/json') >= 0) {
                    data = JSON.parse(request.content);
                } else { //TODO: Let the service declare what content type it supports.
                    response.status = 415;
                    return response.send();
                }
                request.data = data;
            }
            return this[doMethod](request, response);
        };

        var that = this;

        var handleError = function(response, error) {
                log("Failed to access data provider while processing REST request. " + error);
                response.status = 500;
                return response.send();
            };

        /**
         * Method called when servicing a HTTP GET request.
         * The default behavior is to respond with a 404, unless a data provider has been associated with this service.
         * With a data provider, the behavior of this method is to use the declared key to lookup a resource and send
         * a 200 response containing the resource. If the resource wasn't found a 404
         * response will be sent.
         *
         * @param {RestRequest} request The request object received
         * @param {RestResponse} response The response object to send
         * @memberof RestService
         */
        this.onGet = function(request, response) {
            if (this.getDataProvider()) {
                if (this.attributes.key) {
                    var promise = this.getDataProvider().get(this.attributes.key);
                } else {
                    if (request.range) {
                        var count = request.range.count;
                        var offset = request.range.offset;
                    }
                    var promise = this.getDataProvider().getCollection(
                    parseQuery(request.query), count, offset).then(function(result) {
                        response.range = {
                            count: result.count,
                            offset: result.offset,
                            total: result.total
                        };
                        return result.data;
                        // TODO: Set Accept-Ranges header. What range unit to set?
                    });
                }
                return promise.then(function(result) {
                    return response.send(result);
                }, function(error) {
                    return handleError(response, error)
                });
            } else {
                return response.send();
            }
        };

        /**
         * Method called when servicing a HTTP PUT request.
         * The default behavior is to respond with a 404, unless a data provider has been associated with this service.
         * With a data provider, the behavior of this method is to pass the data object in this request to the data provier's
         * put method with the declared key, or putCollection if no key was specified, and send either an empty 204 response
         * or a 200 response containing the return value from the data provider.
         *
         * @param {RestRequest} request The request object received
         * @param {RestResponse} response The response object to send
         * @memberof RestService
         */
        this.onPut = function(request, response) {
            if (this.getDataProvider()) {
                if (this.attributes.key) {
                    var promise = this.getDataProvider().put(this.attributes.key, request.data);
                } else {
                    // Update collection. TODO: Should query params be passed in?
                    var promise = this.getDataProvider().putCollection(request.data);
                }
                return promise.then(function(result) {
                    return response.send(result);
                }, function(error) {
                    return handleError(response, error);
                });
            } else {
                return response.send();
            }
        };

        /**
         * Method called when servicing a HTTP POST request.
         * The default behavior is to respond with a 404, unless a data provider has been associated with this service.
         * With a data provider, the behavior of this method is to pass the data object in this request to the data provier's
         * create method, and send 201 response with a Location header pointing to the resource which ID is the return
         * value from the data provider.
         *
         * @param {RestRequest} request The request object received
         * @param {RestResponse} response The response object to send
         * @memberof RestService
         */
        this.onPost = function(request, response) {
            if (this.getDataProvider()) {
                return this.getDataProvider().create(request.data).then(function(id) {
                    if (id) {
                        response.headers.Location = this.attributes.url + "/" + id;
                    }
                    response.status = 201;
                    return response.send();
                }, function(error) {
                    return handleError(response, error);
                });
            } else {
                return response.send();
            }
        };

        /**
         * Method called when servicing a HTTP DELETE request.
         * The default behavior is to respond with a 404, unless a data provider has been associated with this service.
         * With a data provider, the behavior of this method is to call the data provier's
         * delete method with the declared key, or deleteCollection method if no key is specified,
         * and send either an empty 204 response or a 200 response containing the return
         * value from the data provider.
         *
         * @param {RestRequest} request The request object received
         * @param {RestResponse} response The response object to send
         * @memberof RestService
         */
        this.onDelete = function(request, response) {
            if (this.getDataProvider()) {
                if (this.attributes.key) {
                    var promise = this.getDataProvider().del(this.attributes.key);
                } else {
                    // Delete collection. TODO: Should query params be passed in?
                    var promise = this.getDataProvider().delCollection();
                }
                return promise.then(function(result) {
                    return response.send(result);
                }, function(error) {
                    return handleError(response, error);
                });
            } else {
                return response.send();
            }
        };

        /**
         * Method called when servicing a HTTP OPTIONS request.
         * The default behavior is to respond with a 200 OK.
         *
         * @param {RestRequest} request The request object received
         * @param {RestResponse} response The response object to send
         * @memberof RestService
         */
        this.onOptions = function(request, response) {
            response.status = 200;
            return response.send();
        };

        /**
         * Get the data provider associated with this rest service.
         * @return the data provider
         * @memberof RestService
         */
        this.getDataProvider = function() {
            return this.attributes.dataProvider;
        };

        var parseQuery = function(query) {
                if (query) {
                    var queryObject = {};
                    var params = query.split('&');
                    for (var i in params) {
                        var param = params[i].split('=');
                        if (param[1]) {
                            queryObject[param[0]] = param[1]; // TODO: Do I have to
                            // URL-decode the
                            // parameters?
                        }
                    }
                    return queryObject;
                }
            };
    };
var RestService = new RestService();

exports.RestService = RestService;

/**
 * PushService prototype constructor.
 *
 * @name PushService
 * @class
 * @classdesc The push service is an API for sending message asynchronously using Server Sent Event.
 * It's exposed to the browser as a URL or URL pattern where each unique URL is associated with a {@link PushContext}.
 * The PushContext is used to send messages to the browser using the {@link PushContext.sendMessage}.
 * A PushService can be bound to a {@link avatar/jms~JMS} instance in which case it will listen to the JMS destination
 * and send out an SSE message for each received JMS message.
 * @see {@link module:org/glassfish/avatar.registerPushService registerPushService}
 */
var PushService = function() {

        /**
         * Called when a browser connects to this push service.
         * @param {PushContext} context The {@link PushContext} representing the incoming connection.
         * @param {ServiceRequest} request The request object. Can be used to get custom headers, etc.
         * @param {ServiceResponse} response The response object. Can be used to set custom headers, etc.
         * @memberof PushService
         */
        this.onOpen = function(context) {
            var jmsMeta = this.attributes.jms;
            if (jmsMeta && !context.jms) {
                context.jms = {};
                context.jms.listener = function(message) {
                    if (willLogFine()) logFine("Received JMS message: " + JSON.stringify(message));
                    context.sendMessage(message).then(undefined, function(error) {
                        log("Failed to send message from JMS: " + error);
                    });
                };
                initJMSBinding(this, context.jms.listener, context.jms).then(undefined, function(error) {
                    //peer.close(); TODO: There is no close method on the peer!!
                    log("Failed to create JMS consumer. " + error);
                });
            }
        };

        /**
         * Called when the server has detected that the connection is closed.
         * @param {PushContext} context The context representing the closed connection.
         * @memberof PushService
         */
        this.onClose = function(context) {
            if (context.jms) {
                context.jms.instance.close();
                context.jms = undefined;
            }
        };

        /**
         * This method is called when the timer fires that was set using {@link PushContext.setTimeout}.
         * @param {PushContext} context The context representing the connection.
         * @memberof PushService
         */
        this.onTimeout = function(context) {};

        //TODO: Is this method ever called??
        this.onError = function(error) {};
    };
var PushService = new PushService();

exports.PushService = PushService;

/**
 * SocketService prototype constructor.
 *
 * @name SocketService
 * @class
 * @classdesc The socket service is an API for receiving and sending message over a WebSocket connection.
 * It's exposed to the browser as a URL or URL pattern where each unique URL is associated with a {@link SocketContext}.
 * The context has a set of {@link Peer}s representing each client connected to that specific URL. An implementation
 * of a SocketService can choose to send a message to one single peer using the {@link Peer.send} method or
 * to all peers on the context using the {@link SocketContext.sendAll} method.
 * A SocketService can be bound to a {@link avatar/jms~JMS} instance in which case it will relay all messages to a given JMS
 * destination and vice versa.
 * @see {@link module:org/glassfish/avatar.registerSocketService registerSocketService}
 *
 */
var SocketService = function() {

        /**
         * This method is called when a browser sets up a new WebSocket connection to a URL that matches the configured
         * URL pattern of this socket service. The connection will be represented by a {@link Peer} object that is added to
         * the {@link SocketContext} for this URL. If this is the first connection to this specific URL then a new {@link SocketContext} is created.
         * If a {@link module:org/glassfish/avatar/jms.JMS JMS} instance is bound to this SocketService, a listener will be added to the JMS instance which will
         * send out the received JMS message to all peers on the context.
         * @param peer The {@link Peer} object representing the client
         * @memberof SocketService
         */
        this.onOpen = function(peer) {
            var ctx = peer.getContext();
            var jmsMeta = this.attributes.jms;
            if (jmsMeta && !ctx.jms) {
                ctx.jms = {};
                ctx.jms.listener = function(message) {
                    ctx.sendAll(message).then(undefined, function(error) {
                        log("Failed to send message from JMS: " + error);
                    });
                };
                initJMSBinding(this, ctx.jms.listener, ctx.jms).then(undefined, function(error) {
                    //peer.close(); TODO: There is no close method on the peer!!
                    log("Failed to create JMS consumer. " + error);
                });
                var messageProperties = jmsMeta.messageProperties;
                if (messageProperties) {
                    //                ctx.jms.properties = {};
                    ctx.jms.properties = messageProperties;
                    //                for (p in messageProperties) {
                    //                    ctx.jms.properties[p] = messageProperties[p].call(this);
                    //                }
                }
            }
        };

        /**
         * This method is called when the server has detected that a connection with a client has been closed.
         * @param peer The {@link Peer} object representing the client connection that was closed
         * @memberof SocketService
         */
        this.onClose = function(peer) {
            var ctx = peer.getContext();
            if (ctx.jms && ctx.getPeers().length < 1) {
                var instance = ctx.jms.instance;
                ctx.jms = undefined;
                instance.close();
            }
        };


        /**
         * Called when a message has been received.
         * @param peer the {@link Peer} object representing the sender of the message
         * @param message the message. This is a string representation of the data that was sent.
         * @memberof SocketService
         */
        this.onMessage = function(peer, message) {
            if (this.attributes.jms) {
                var ctx = peer.getContext();
                ctx.jms.instance.send(message, ctx.jms.properties).then(undefined, function(error) {
                    log("Failed to put message on JMS queue: " + error);
                });
            }
        };

        /**
         * Called when there is a i/o error
         * @param session the {@link Session} object representing the sender of the message
         * @param error the {@link Throwable} representing the error
         * @memberof SocketService
         */
        this.onError = function(session, error) {};

    };
var SocketService = new SocketService();

exports.SocketService = SocketService;

/**
 * Registers a service.
 *
 * @param {object} service An object of class {@link RestService}, {@link PushService}, or {@link SocketService}.
 * @static
 */
var register = function(service) {
        // Make sure that our container is configured to accept requests.
        if (serviceContainer.acceptsRequests()) {
            var url = service.attributes.url;
            if (registry.hasUrl(url)) {
                throw "URL " + url + " is in use by another service.";
            }
            service.registry[url] = service;
            compileMetaData(service.attributes);
        } else {
            throw 'This thread does not process service requests.';
        }
    };

exports.register = register;

/**
 * Meta data for a rest service.
 * @typedef RestMetaData
 * @property {string} url An URL patters identifying this rest resource. The pattern can contain wildcards (*) and parameters (e.g. {name}).
 * The URL is relative to the applications context root.
 * @property {array} methods (optional) An array of all supported methods. Any request of a method not in this list will be rejected with a 403.
 * @property {object} dataProvider (optional) A data provider that stores the resources exposed by this rest service.
 * @property {string} key (optional) The unique key for a resource exposed by this rest service. The key can contain contain an EL expression
 * that references properties on the request and parameters on the URL pattern.
 * @property {CorsMetaData} cors (optional) the CORS policy for this service. If not provided cross-origin requests will not be allowed.
 * @property {AuthMetaData} authorization (optional) a set of authorization rules to protect the service from unauthorized access.
 * The absence of this property means that any request is allowed.
 */

/**
 * Creates a {@link RestService} instance. The returned object must be passed in to the {@link module:org/glassfish/avatar.register register} function
 * before it can be used.
 * @param {RestMetaData} metadata The metadata for the service, e.g. {url: 'data/myservice'}.
 * @param {function} restService The REST service constructor.
 * @static
 */
var restService = function(metadata, restService) {
        restService.prototype = this.RestService;
        restService.attributes = metadata;
        validateCORSRules(metadata.cors);
        restService.registry = registry.restServices;
        return restService;
    };

exports.restService = restService;


/**
 * Create and register a {@link RestService} instance. The registered REST service will have {@link RestService}
 * as a prototype and may override any of its methods to add business logic or rely the default behavior.
 * @param {RestMetaData} metadata The metadata for the service, e.g. {url: 'data/myservice'}
 * @param {function} restService The REST service constructor.
 * @static
 *
 * @example <caption>A simple REST service that returns a hello world object for any GET request.
 * Any other method will be handled by the {@link RestService} prototype, which in this case is
 * to return a 404 Not Found.</caption>
 *
 * var avatar = require("org/glassfish/avatar");
 *
 * avatar.registerRestService({
 *     url: "/data/helloworld",
 *     methods: ["GET"]},
 *   function(){
 *     this.onGet = function(req, res){
 *       res.send({message: "Hello World!"});
 *     }
 *   }
 * );
 *
 * @example <caption>This example shows how a RestService can be bound to a data provider
 * for automatic population of data.</caption>
 *
 * var avatar = require("org/glassfish/avatar");
 * var myDataProvider = new avatar.FileDataProvider({
 *     filename : "cars.txt",
 *     key : "id"});
 *
 * avatar.registerRestService(
 *   {url:"data/cars/{carid}", //Capture the last path segment as the resource id
 *   dataProvider:myDataProvider,
 *   key:"#{this.carid}"}, //Use the captured resource id as key for the data provider
 *   function(){} //The RestService prototype populates the data from the configured data provider.
 * );
 *
 *
 */
var registerRestService = function(metadata, restService) {
        register(this.restService(metadata, restService));
    };

exports.registerRestService = registerRestService;

/**
 * Meta data for a JMS binding. A JMS instance can be bound to push and socket services.
 * @typedef JMSMetaData
 * @property {string} connectionFactoryName The name of the system configured  javax.jms.ConnectionFactory to use
 * for connecting to a JMS Provider.
 * @property {string} destinationName A system configured Queue or Topic that will be used to send and receive
 * messages.
 * @property {string} messageSelector (optional) A JMS message selector as defined in the
 * {@link http://download.oracle.com/otndocs/jcp/jms-2_0-pfd-spec/index.html JMS specification}.
 * @property {string} messageProperties (optional) An object that will be used to populate message properties
 * to each JMS message being sent. Each property on this object will be added to the message according to the
 * rules of the {@link http://download.oracle.com/otndocs/jcp/jms-2_0-pfd-spec/index.html JMS specification}.
 * This property is not applicable for a binding with a {@link PushService}.
 * @property {object} instance A {@link module:org/glassfish/avatar/jms.JMS JMS} instance that will be bound to this service. If this property is
 * set connectionFactoryName and destinationName will be ignored.
 */

var initJMSBinding = function(service, listener, context) {
        var metadata = service.attributes.jms;
        var messageSelector = metadata.messageSelector;
        var instance = metadata.instance || new jms.JMS({
            connectionFactoryName: metadata.connectionFactoryName,
            destinationName: metadata.destinationName,
            messageSelector: messageSelector
        });
        context.instance = instance;
        return instance.addListener(listener)
    };

var validateCORSRules = function(cors) {
        if (cors) {
            if (cors.allowOrigin) {
                if (cors.allowOrigin !== '*') {
                    if (typeof cors.allowOrigin === 'string') {
                        cors.allowOrigin = [cors.allowOrigin];
                    } else if (!Array.isArray(cors.allowOrigin)) {
                        throw 'cors.allowOrigin must be either the allowed origin URI as a string, the wild card string "*", or an array of allowed origin URIs.';
                    }
                    cors.allowOrigin = cors.allowOrigin.map(function(allowOrigin) {
                        //var origin = url.parse(callowOrigin);
                        //The url.parse function doesn't parse URLs without scheme correctly (e.g. "example.com:80")
                        var origin = internal.parseOrigin(allowOrigin);
                        if (!origin.protocol) {
                            origin.protocol = 'http:';
                        }
                        if (!origin.port) {
                            origin.port = '80';
                        }
                        return origin.protocol + '//' + origin.host + ':' + origin.port;
                    });
                }

            }
            if (cors.allowHeaders) {
                if (typeof cors.allowHeaders === 'string') {
                    cors.allowHeaders = [cors.allowHeaders];
                } else if (!Array.isArray(cors.allowHeaders)) {
                    throw 'cors.allowHeaders must be either the allowed header name as a string, or an array of allowed header names.';
                }
                cors.allowHeaders = cors.allowHeaders.map(function(header) {
                    return header.toLowerCase();
                });
            }
            if (cors.allowMethods) {
                if (typeof cors.allowMethods === 'string') {
                    cors.allowMethods = [cors.allowMethods];
                } else if (!Array.isArray(cors.allowMethods)) {
                    throw 'cors.allowMethods must be either the allowed method name as a string, or an array of allowed method names.';
                }
            }
        }
    };

/**
 * Meta data for a {@link PushService}.
 * @typedef PushServiceMetaData
 * @property {string} url An URL patters identifying this push service. The pattern can contain wildcards (*) and parameters (e.g. {name}).
 * The URL is relative to the applications context root.
 * @property {JMSMetaData} jms (optional) Meta data to bind this push service to a {@link module:org/glassfish/avatar/jms.JMS JMS} destination.
 * @property {CorsMetaData} cors (optional) the CORS policy for this service. If not provided cross-origin requests will not be allowed.
 * @property {AuthMetaData} authorization (optional) a set of authorization rules to protect the service from unauthorized access.
 * The absence of this property means that any request is allowed.
 */

/**
 * Create a Server-Sent Event {@link PushService} instance.
 * @param {PushServiceMetaData} metadata The metadata, e.g. {url: 'push/myservice'}
 * @param {function} pushService The push service constructor. An instance will be created with
 * this constructor having {@link PushService} as prototype.
 * @static
 */
var pushService = function(metadata, pushService) {
        pushService.prototype = this.PushService;
        pushService.attributes = metadata;
        validateCORSRules(metadata.cors);
        pushService.registry = registry.pushServices;
        return pushService;
    };

exports.pushService = pushService;

/**
 * Create and register a Server-Sent Event {@link PushService} instance.
 * @param {PushServiceMetaData} metadata The metadata, e.g. {url: 'push/myservice'}
 * @param {function} pushService The push service constructor. An instance will be created with
 * this constructor having {@link PushService} as prototype.
 * @static
 *
 * @example <caption>A simple push service that sends a single message after 5 seconds.</caption>
 * avatar.registerPushService({
 *   url: "push/stocks"
 * }, function() {
 *
 *   this.onOpen = function(context){
 *     context.setTimeout(5000); //Start a timer for 5 seconds
 *   };
 *
 *   this.onTimeout = function(context) {
 *     context.sendMessage("HelloWorld!"); //When timer fires send an SSE message.
 *   };
 * });
 *
 * @example <caption> Here is an example of a push service that listens for log messages
 * on a JMS Queue and sends then out to the browser.</caption>
 *
 * avatar.registerPushService({url: "/push/chat",
 *   jms: {
 *     connectionFactoryName: "jms/MyConnectionFactory",
 *     destinationName: "jms/LogQueue"}
 *   }, function(){});
 */
var registerPushService = function(metadata, pushService) {
        register(this.pushService(metadata, pushService));
    };

exports.registerPushService = registerPushService;

/**
 * Meta data for a {@link SocketService}.
 * @typedef SocketServiceMetaData
 * @property {string} url An URL patters identifying this websocket service. The pattern can contain wildcards (*) and parameters (e.g. {name}).
 * The URL is relative to the applications context root.
 * @property {JMSMetaData} jms (optional) Meta data to bind this socket service to a {@link module:org/glassfish/avatar/jms.JMS JMS} destination.
 */

/**
 * Create a WebSocket {@link SocketService} instance. This instance has to be registered using the
 * {@link module:org/glassfish/avatar.register register} function.
 * @param {SocketServiceMetaData} metadata The metadata, e.g. {url: 'websocket/myservice'}
 * @param {function} socketService The socket service constructor. An instance will be created with
 * this constructor having {@link SocketService} as prototype.
 * @static
 */
var socketService = function(metadata, socketService) {
        socketService.prototype = this.SocketService;
        socketService.attributes = metadata;
        socketService.registry = registry.socketServices;
        return socketService;
    };

exports.socketService = socketService;

/**
 * Creates and registers a WebSocket {@link SocketService} instance.
 * @param {SocketServiceMetaData} metadata The metadata, e.g. {url: 'websocket/myservice'}
 * @param {function} socketService The socket service constructor. An instance will be created with
 * this constructor having {@link SocketService} as prototype.
 * @static
 *
 * @example <caption>This example shows how to use a socket service to implement a simple chat service</caption>
 *
 * var avatar = require("org/glassfish/avatar");
 *
 * avatar.registerSocketService({
 *   url: "/websocket/chat"
 * }, function() {
 *   this.onMessage = function(peer, message){
 *     //Forward the message to all peers.
 *     peer.getContext().sendAll(message);
 *   };
 * });
 *
 * @example <caption>An example that shows how to use a binding to JMS to implement a chat service with
 * multiple chat rooms. </caption>
 *
 * var avatar = require("org/glassfish/avatar");
 *
 * avatar.registerSocketService({
 *    url: "/websocket/jmschat/{chatroom}",
 *    jms: {
 *      connectionFactoryName: "jms/MyConnectionFactory",
 *      destinationName: "jms/ChatTopic",
 *      //A property on the message is used to dispatch to different chat rooms.
 *      messageSelector: "chatroom='#{this.chatroom}'",
 *      messageProperties: {chatroom: "#{this.chatroom}"}}},
 *  function() {});//The prototype of the socket service takes care of passing messages to and from JMS
 *
 *
 */
var registerSocketService = function(metadata, socketService) {
        register(this.socketService(metadata, socketService));
    };

exports.registerSocketService = registerSocketService;

var factory = Java.type('org.glassfish.avatar.runtime.service.inject.InjectorFactory').factory();
var InjectorArray = Java.type("org.glassfish.avatar.runtime.service.inject.Injector[]");

var compileInjectors = function (inject) {
    if (inject) {
        var keys = Object.keys(inject);
        var length = keys.length;
        var injectors = new InjectorArray(length);
        for (var i = 0; i < length; i++) {
            var key = keys[i];
            var injector;
            var value = inject[key];
            if (typeof value === 'string') {
                injector = factory.injectorFor({key: key, type: value});
            } else if (Array.isArray(value)) {
                var type = value[0];
                var qualifiers = value.slice(1);
                injector = factory.injectorFor({key: key, type: type, qualifiers: qualifiers});
            } else {
                throw 'unknown injection type for ' + key + ' in ' + JSON.stringify(inject);
            }
            injectors[i] = injector;
            inject[key] = injector; // So we can find by name
        }
        return injectors;
    }
    return null;
};

/*
 * Compile any injectors or EL expressions in the metadata recursively. Objects that have an
 * ELCompilerIgnore property will be ignored.
 */
var compileMetaData = function (metadata) {
    for (var p in metadata) {
        if (p === 'inject') {
            metadata.injectors = compileInjectors(metadata[p]);
        } else {
            if (typeof metadata[p] === 'string') {
                metadata[p] = compileExpression(metadata[p]);
            } else if (typeof metadata[p] === 'object' && !metadata[p].ELCompilerIgnore) {
                compileMetaData(metadata[p]);
            }
        }
    }
};

exports.compileMetaData = compileMetaData; // test only
/*
 * A very basic EL compiler that can handle expressions like #{this.a} or
 * #{request.b}.
 */
var compileExpression = function(expression) {
        var start = expression.indexOf("#{");
        if (start >= 0) {
            var end = expression.indexOf("}");
            if (end > start) {
                var exp = expression.substring(start + 2, end);
                if (exp.indexOf("this.") === 0) {
                    return function() {
                        return expression.substring(0, start) + this[exp.substring(5)]
                                + expression.substring(end + 1, expression.length);
                    };
                } else if (exp.indexOf("request.") === 0) {
                    return function(request) {
                        return expression.substring(0, start) + request[exp.substring(8)]
                                + expression.substring(end + 1, expression.length);
                    };
                }
            }
        } else {
            return expression;
        }
    };

exports.compileExpression = compileExpression; // test only
exports.FileDataProvider = dataproviders.FileDataProvider;
exports.JPADataProvider = dataproviders.JPADataProvider;
exports.NoSqlDataProvider = dataproviders.NoSqlDataProvider;
exports.JMS = jms.JMS;

/**
 * Registers a shutdown hook declared in the service module
 * @param hook  the shutdown hook to execute
 */
exports.registerShutdownHook = function(hook) {
    registry.shutdownHook = hook;
};

/**
 * Log a message.
 * @param {string} msg The message.
 * @function
 */
exports.log = common.log; // re-export for convenience.
/**
 * Log a warning message.
 * @param {string} msg The message.
 * @param {object} exception The exception, if any.
 * @function
 */
exports.logWarning = common.logWarning; // re-export for convenience.
/**
 * Returns true if FINE logging is enabled.
 * @returns {boolean}
 * @function
 */
exports.willLogFine = common.willLogFine; // re-export for convenience.
/**
 * Log a message if logger is at level FINE or below.
 * @param {string} msg The message.
 * @function
 */
exports.logFine = common.logFine; // re-export for convenience.
/**
 * The {@link module:org/glassfish/avatar/common.application application} instance.
 */
exports.application = common.application; // Re-export for convenience

// Do this last!
require('process').nextTick(function() {
    //This guarantees that node doesn't shut down when it receives an uncaught exception.
    process.on('uncaughtException', function(error) {
        common.logWarning('Uncaught exception', error);
    });
    serviceContainer.startupCompleted();
});
