/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2013-2014 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of the GNU General
 * Public License Version 2 only ("GPL"). You may not use this file except
 * in compliance with the License.  You can obtain a copy of the License at
 * https://avatar.java.net/license.html or legal/LICENSE.txt.  See the
 * License for the specific language governing permissions and limitations
 * under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 */

/**
 * This module contains implementations of the asynchronous data provider API.
 * This API is designed to make it easy to persist and access REST resources in different
 * platform specific storages. A data provider consists of the methods get, getCollection,
 * put, putCollection, del, delCollection, and create. These methods have a
 * well defined mapping to the HTTP methods GET, PUT, POST, and DELETE as described in
 * {@link RestService}.
 * @module org/glassfish/avatar/dataproviders
 */

'use strict';

var events = require("events");
var util = require("util");
var common = require('org/glassfish/avatar/common');
var Promise = common.Promise;
var wrapCallback = common.wrapCallback;
var nodeCall = common.nodeCall;


//The data base listener support is an optional feature. It not available use the EventEmitter.
var listenerSupport = common.optionalModule('org/glassfish/avatar/dp-listener',
    'Database change notification support is not available.',
    events.EventEmitter);


/**
 * Callback for the get method of a data provider.
 * @callback getCallback
 * @param {string} error If an error occurred, otherwise null
 * @param {object} item the item or undefined if no item was found.
 */

/**
 * The result of a getCollection call on a data provider.
 * @typedef GetCollectionResult
 * @property data {array} the collection or a subset of it in case range parameters were passed in.
 * @property count {number} (optional) the number of items in the collection returned if a subset was
 * requested using the count parameter
 * @property offset {number} (optional) the offset in the total collection that this sub set represents
 * if an offset parameter was passed to getCollection.
 * @property total {number} (optional) the total number of the collection stored if a count parameter
 * was passed to getCollection.
 */

/**
 * Callback for the getCollection method of a data provider.
 * @callback getCollectionCallback
 * @param {string} error If an error occurred, otherwise null
 * @param {GetCollectionResult} result an object
 */

/**
 * Callback for the create method of a data provider.
 * @callback createCallback
 * @param {string} error If an error occurred, otherwise null
 * @param {string} id the generated id for the created item
 */

var FileDataProviderHelper = Java.type('org.glassfish.avatar.runtime.service.FileDataProviderHelper');

/**
 * Constructor for creating a FileDataProvider that stores resources in the specified file
 * using the specified key.
 * @param {object} properties Meta data to the FileDataProvider that has to contain the
 * properties "filename" and "key".
 *
 * @classdesc A data provider that stores resource entities in a file on the file system.
 * This data provider is not intended to be used in production environments.
 * Each data provider instance stores one collection of resources where each
 * resource entity is identified by a unique key.
 * The data provider implements the EventEmitter interface for notifying on changes to
 * stored entities. Note that these events are only emitted to listeners in the same
 * process that did the change so in a distributed environment some other means (such as
 * JMS) is needed to propagate these changes across the cluster.
 *
 * @example
 * var dp = new avatar.FileDataProvider({
 *   filename: "books.txt",
 *   key: "id"
 * });
 *
 * @class
 * @static
 */
var FileDataProvider = function (properties) {
    this.ELCompilerIgnore = true;
    this.filename = properties.filename;
    this.key = properties.key;
    var that = this;
    var helper = new FileDataProviderHelper(serviceContainer);

    /**
     * Method for retrieving the entire collection of resources that this
     * data provider stores.
     * @param {getCollectionCallback} callback (optional) callback function to be called with the result.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.getCollection = function (callback) {
        var deferred = Promise.defer();
        helper.getCollection(this.filename, function (event, args) {
            deferred.resolve({
                'data': JSON.parse(args[0])
            });
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback((typeof callback === 'function') && callback, deferred.promise);
    };

    /**
     * Method for retrieving one single resource entity based on a passed in key.
     * @param {string} key The unique key for the resource to get.
     * @param {getCallback} callback (optional) callback function to be called with the result.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.get = function (id, callback) {
        var deferred = Promise.defer();
        helper.get(this.filename, id, function (event, args) {
            if (args && args.length > 0) {
                deferred.resolve(JSON.parse(args[0]));
            } else {
                deferred.resolve();
            }
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    /**
     * Method for storing a single resource entity.
     * @param {string} key The unique key for the resource to get.
     * @param {object} item The resource entity to store.
     * @param {noargCallback} callback (optional) callback function to be called when operation has completed.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     * @fires itemUpdated
     */
    this.put = function (id, item, callback) {
        var deferred = Promise.defer();
        helper.put(this.filename, id, JSON.stringify(item), function (event, args) {
            /**
             * Indicates that an item has been updated
             * @event itemUpdated
             * @property {string} id The id of the updated item
             * @property {object} item the new item value
             */
            that.emit("itemUpdated", id, item);
            deferred.resolve(null);
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    /**
     * Method for storing the entire resource collection.
     * This method will completely replace the existing collection with the new one.
     * @param {array} items The resource collection to store.
     * @param {noargCallback} callback (optional) callback function to be called when operation has completed.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     * @fires collectionUpdated
     */
    this.putCollection = function (items, callback) {
        var deferred = Promise.defer();
        helper.putCollection(this.filename, Java.to(items.map(function (item) {
            return 'id: ' + item[that.key] + ':' + JSON.stringify(item);
        }), 'java.util.List'), function (event, args) {
            /**
             * Indicates that the whole collection of items has been updated.
             * @event collectionUpdated
             * @property {array} items the new collection of items.
             */
            that.emit("collectionUpdated", items);
            deferred.resolve(null);
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    /**
     * Method for deleting a resource entity.
     * @param {string} id The unique id of the resource to delete.
     * @param {noargCallback} callback (optional) callback function to be called when operation has completed.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     * @fires itemDeleted
     */
    this.del = function (id, callback) {
        var deferred = Promise.defer();
        helper.del(this.filename, id, function (event, args) {
            if (args[0]) {
                /**
                 * Indicates that an item has been deleted.
                 * @event itemDeleted
                 * @property {string} id the id of the deleted item
                 */
                that.emit("itemDeleted", id);
                deferred.resolve(null);
            } else {
                deferred.resolve();
            }
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };


    /**
     * Method for deleting the entire resource collection.
     * @param {noargCallback} callback (optional) callback function to be called when operation has completed.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     * @fires collectionDeleted
     */
    this.delCollection = function (callback) {
        var deferred = Promise.defer();
        helper.delCollection(this.filename, function (event, args) {
            /**
             * Indicates that the whole collection of items has been deleted.
             * @event collectionDeleted
             */
            that.emit("collectionDeleted");
            deferred.resolve(null);
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    var generateId = function () {
        return Math.floor((Math.random() * 1000000000) + 1); //pseudo-random integer between 1 to one billion.
    };

    /**
     * Method for creating a resource entity. An id will be generated for the item and passed in to
     * the callback or promise resolution.
     * @param {object} item the item to create
     * @param {createCallback} callback (optional) callback function to be called when operation has completed.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     * @fires itemUpdated
     */
    this.create = function (item, callback) {
        var deferred = Promise.defer();
        var id = generateId();
        item[this.key] = id;
        helper.create(this.filename, JSON.stringify(item), id, function (event, args) {
            /**
             * Indicates that an item has been updated
             * @event itemUpdated
             * @property {string} id The id of the updated item
             * @property {object} item the new item value
             */
            that.emit("itemUpdated", id, item);
            deferred.resolve(id);
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };
};

util.inherits(FileDataProvider, events.EventEmitter);
exports.FileDataProvider = FileDataProvider;

/**
 * @typedef JPADataProviderMetaData
 * Meta data for the JPA data provider.
 * @property {string} persistenceUnit The persistence unit that has been configured in a platform
 * specific way. Typically in a persistence.xml.
 * @property {string} entityType The type of the JPA entity represented by this data provider.
 * @property {boolean} createTables (optional) If true, data base tables will be created on first
 * access unless they already exists. Note! This property only works with certain JPA implementations.
 * Intended for development only.
 */

/**
 * Constructor for creating a JPADataProvider.
 * @param {JPADataProviderMetaData} properties Meta data to the JPADataProvider that has to contain the
 * properties "filename" and "key".
 *
 * @classdesc A data provider that stores resource items as JPA entity beans, typically persisted in a database.
 * Each data provider instance represents one JPA entity type and that type has to be available in the system
 * (typically defined in an orm.xml or a Java class with annotations).
 * The data provider implements the EventEmitter interface for notifying on changes to
 * stored entities.
 * When using an Oracle data base as persistent storage this events will be emitted by data providers in other
 * processes that are connected to the same data base. For other storages the event is limited to in-process listeners.
 * Note! With an Oracle DB as storage the {@link itemDeleted} event will not contain the id of the deleted item
 * due to limitations in the underlying API.
 *
 * <p><b>Note that <tt>JPADataProvider</tt> is supported only in Avatar GF</b>.
 *
 * @example
 * var myDataProvider = new avatar.JPADataProvider(
 *   {persistenceUnit:"test",
 *   createTables:"true",
 *   entityType:"Foo"}
 * );
 *
 *
 * @class
 * @static
 */
var JPADataProvider = function (properties, content) {
    this.ELCompilerIgnore = true;
    this.persistenceUnit = properties.persistenceUnit;
    this.entityType = properties.entityType;
    if (properties.dataModel) {
        this.dataModel = properties.dataModel;
    }
    var that = this;
    var initHelperEmitter;
    var initializingHelper = false;
    var factory = Packages.org.glassfish.avatar.runtime.service.jpa.JPADataProviderHelper.newFactory(serviceContainer);

    var initHelper = function () {
        if (!initHelperEmitter) {
            initHelperEmitter = new events.EventEmitter();
        }
        if (!initializingHelper) {
            initializingHelper = true;
            factory.getHelper(that.persistenceUnit, properties.createTables, that.dataModel != undefined, function (name, args) {
                that.helper = args[0];
                initHelperEmitter.emit("initialized");
                initializingHelper = false;
            }, function (error) {
                initHelperEmitter.emit("error", error);
                initializingHelper = false;
            });
        }
        return initHelperEmitter;
    };

    var getHelper = function () {
        var deferred = Promise.defer();
        initHelper().once("initialized",function () {
            deferred.resolve(that.helper);
        }).once("error", function (error) {
                deferred.reject(error);
            });
        return deferred.promise;
    }

    if (this.dataModel) {
        var doGetCollection = function (helper, parameters, count, offset, callback) {
            helper.getCollection(moduleName, that.dataModel.name, that.dataModel.key, that.dataModel.fields, parameters, count, offset, function (name, args) {
                callback(null, args[0]);
            });
        };
    } else {
        var doGetCollection = function (helper, parameters, count, offset, callback) {
            helper.getCollection(that.entityType, parameters, count, offset, function (name, args) {
                callback.apply(null, args);
            });
        };
    }

    /**
     * Method for retrieving the entire collection of resources that this
     * data provider stores.
     * @param {object} parameters (optional) query parameters for getting a subset of the collection. Each
     * property of the passed in object is treated as a parameter where the name of the property corresponds
     * to a field of the entity and the value corresponds to the value of that entity field.
     * @param {number} count (optional) the maximum number of entities in the result
     * @param {number} offset (optional) the lowest index to include in the result set.
     * @param {getCollectionCallback} callback (optional) callback function to be called with the result.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.getCollection = function (parameters, count, offset, callback) {
        var lastarg = arguments[arguments.length - 1];
        if (typeof lastarg === 'function') {
            callback = lastarg;
        }
        ;
        return wrapCallback(callback, Promise(this.helper || getHelper()).then(function (helper) {
            return nodeCall(doGetCollection, helper, parameters, count, offset);
        }).then(function (result) {
                var data = that.toJSObject(result.getResult());
                // Return value {data, count, offset, total}
                return {
                    "data": data,
                    "count": result.getCount(),
                    "offset": result.getOffset(),
                    "total": result.getTotalCount()
                };
            }));
    };

    if (this.dataModel) {
        var doGet = function (helper, id, callback) {
            helper.get(moduleName, that.dataModel.name, that.dataModel.key, id, that.dataModel.fields, function (name, args) {
                callback(null, args[0]);
            });
        };
    } else {
        var doGet = function (helper, id, callback) {
            helper.get(that.entityType, id, function (name, args) {
                callback.apply(null, args);
            });
        };
    }

    /**
     * Method for retrieving one single resource entity based on a passed in key.
     * @param {string} key The unique key for the resource to get.
     * @param {getCallback} callback (optional) callback function to be called with the result.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.get = function (id, callback) {
        return wrapCallback(callback, Promise(this.helper || getHelper()).then(function (helper) {
            return nodeCall(doGet, helper, id);
        }).then(function (result) {
                return that.toJSObject(result) || undefined; //Make sure null isn't sent back
            }));
    };

    if (this.dataModel) {
        var doPut = function (helper, id, item, callback) {
            helper.put(moduleName, that.dataModel.name, that.dataModel.key, id, item, that.dataModel.fields, function (name, args) {
                callback(null, args[0]);
            });
        };
    } else {
        var doPut = function (helper, id, item, callback) {
            helper.put(that.entityType, id, item, function (name, args) {
                callback.apply(null, args);
            });
        };
    }

    /**
     * Method for storing a single resource entity.
     * @param {string} key The unique key for the resource to get.
     * @param {object} item The resource entity to store.
     * @param {noargCallback} callback (optional) callback function to be called when operation has completed.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     * @fires itemUpdated
     */
    this.put = function (id, item, callback) {
        return wrapCallback(callback, Promise(this.helper || getHelper()).then(function (helper) {
            return nodeCall(doPut, helper, id, item);
        }).then(function () {
                that.emit("itemUpdated", id, item);
                return null;
            }));
    };

    if (this.dataModel) {
        var doPutCollection = function (helper, items, callback) {
            helper.putCollection(moduleName, that.dataModel.name, that.dataModel.key, items, that.dataModel.fields, function (name, args) {
                callback(null, args[0]);
            });
        };
    } else {
        var doPutCollection = function (helper, items, callback) {
            helper.putCollection(that.entityType, items, function (name, args) {
                callback.apply(null, args);
            });
        };
    }

    /**
     * Method for storing the entire resource collection.
     * This method will completely replace the existing collection with the new one.
     * @param {array} items The resource collection to store.
     * @param {noargCallback} callback (optional) callback function to be called when operation has completed.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     * @fires collectionUpdated
     */
    this.putCollection = function (items, callback) {
        return wrapCallback(callback, Promise(this.helper || getHelper()).then(function (helper) {
            return nodeCall(doPutCollection, helper, items);
        }).then(function () {
                that.emit("collectionUpdated", items);
                return null; //Successful
            }));
    };

    if (this.dataModel) {
        var doCreate = function (helper, item, callback) {
            helper.create(moduleName, that.dataModel.name, that.dataModel.key, item, that.dataModel.fields, function (name, args) {
                callback(null, args[0]);
            });
        };
    } else {
        var doCreate = function (helper, item, callback) {
            helper.create(that.entityType, item, function (name, args) {
                callback.apply(null, args);
            });
        };
    }

    /**
     * Method for creating a resource entity. An id will be generated for the item and passed in to
     * the callback or promise resolution. The id generation is controlled by the underlying JPA layer
     * so the corresponding entity definition has to have id generation enabled for the primary key
     * of this entity.
     * @param {object} item the item to create
     * @param {createCallback} callback (optional) callback function to be called when operation has completed.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     * @fires itemUpdated
     */
    this.create = function (item, callback) {
        return wrapCallback(callback, Promise(this.helper || getHelper()).then(function (helper) {
            return nodeCall(doCreate, helper, item);
        }).then(function (id) {
                that.emit("itemUpdated", id, item);
                return id; //Returns id of created item
            }));
    }

    if (this.dataModel) {
        var doDelete = function (helper, id, callback) {
            helper.del(moduleName, that.dataModel.name, that.dataModel.key, id, that.dataModel.fields, function (name, args) {
                callback(null, args[0]);
            });
        };
    } else {
        var doDelete = function (helper, id, callback) {
            helper.del(that.entityType, id, function (name, args) {
                callback.apply(null, args);
            });
        };
    }

    /**
     * Method for deleting a resource entity.
     * @param {string} id The unique id of the resource to delete.
     * @param {noargCallback} callback (optional) callback function to be called when operation has completed.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     * @fires itemDeleted
     */
    this.del = function (id, callback) {
        return wrapCallback(callback, Promise(this.helper || getHelper()).then(function (helper) {
            return nodeCall(doDelete, helper, id)
        }).then(function (deleted) {
                that.emit("itemDeleted", id);
                if (deleted) {
                    return null; //Successful
                } else {
                    return undefined; //Not found
                }
            }));
    }

    if (this.dataModel) {
        var doDeleteCollection = function (helper, callback) {
            helper.delCollection(moduleName, that.dataModel.name, that.dataModel.key, that.dataModel.fields, function (name, args) {
                callback(null, args[0]);
            });
        };
    } else {
        var doDeleteCollection = function (helper, callback) {
            helper.delCollection(that.entityType, function (name, args) {
                callback.apply(null, args);
            });
        };
    }

    /**
     * Method for deleting the entire resource collection.
     * @param {noargCallback} callback (optional) callback function to be called when operation has completed.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     * @fires collectionDeleted
     */
    this.delCollection = function (callback) {
        return wrapCallback(callback, Promise(this.helper || getHelper()).then(function (helper) {
            return nodeCall(doDeleteCollection, helper)
        }).then(function () {
                that.emit("collectionDeleted");
                return null; //Successful
            }));
    };

    this.getKey = function (item, type) {
        return item[type.getDescriptor().getPrimaryKeyFieldNames().firstElement()];
    };

    this.toJSObject = function (object) {
        if (object === null) {
            return null;
        } else if (object.getClass().isArray()) {
            var jsObject = [];
            for (var i = 0; i < object.length; i++) {
                jsObject.push(this.toJSObject(object[i]));
            }
            return jsObject;
        } else if (java.util.Map.class.isAssignableFrom(object.getClass())) {
            var jsObject = {};
            var propIter = object.keySet().iterator();
            while (propIter.hasNext()) {
                var prop = propIter.next();
                jsObject[prop] = this.toJSObject(object.get(prop));
            }
            return jsObject;
        } else { // Primitives
            return object;
        }
    };
};

util.inherits(JPADataProvider, listenerSupport);
exports.JPADataProvider = JPADataProvider;

var NoSqlDataProviderHelper = Java.type('org.glassfish.avatar.runtime.service.nosql.NoSqlDataProviderHelper');
var JavaString = Java.type('java.lang.String');

/**
 * Constructor for creating a NoSqlDataProvider that stores values in an Oracle's
 * NoSQL key-value store.
 *
 * @param {object} properties An object containing the properties needed to
 * establish a connection with NoSQL store. These are: storeName, hostName
 * (default "localhost") and port (default "5000").
 *
 * @classdesc A data provider that stores key-value pairs in an Oracle's
 * NoSql data store. This data provider implements the EventEmitter interface for
 * notifying on changes to stored values. Note that these events are only emitted
 * to listeners in the same process that did the change so in a distributed environment
 * some other means (such as JMS) is needed to propagate these changes across the cluster.
 *
 * @example
 * var dp = new avatar.NoDataProvider({
 *   storeName: "myStore",
 *   hostName: "localhost",
 *   port: 5000
 * });
 *
 * @class
 * @static
 */
var NoSqlDataProvider = function (properties) {
    var that = this;
    var helper = new NoSqlDataProviderHelper(
        serviceContainer, properties.storeName, properties.hostName, properties.port);

    this.get = function (key, callback) {
        var deferred = Promise.defer();
        helper.get(key, function (event, args) {
            if (args && args.length > 0) {
                deferred.resolve(JSON.parse(new JavaString(args[0], "utf-8")));
            } else {
                deferred.resolve();
            }
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    this.put = function (key, item, callback) {
        var deferred = Promise.defer();
        helper.put(key, new JavaString(JSON.stringify(item)).getBytes("utf-8"), function (event, args) {
            /**
             * Indicates that an item has been updated
             * @event itemUpdated
             * @property {string} key The key of the updated item
             * @property {object} item the new item value
             */
            that.emit("itemUpdated", key, item);
            deferred.resolve(null);
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    this.del = function (key, callback) {
        var deferred = Promise.defer();
        helper.del(key, function (event, args) {
            if (args[0]) {
                /**
                 * Indicates that an item has been deleted.
                 * @event itemDeleted
                 * @property {string} key the key of the deleted item
                 */
                that.emit("itemDeleted", key);
                deferred.resolve(null);
            } else {
                deferred.resolve();
            }
        }, function (event, args) {
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };
};

util.inherits(NoSqlDataProvider, events.EventEmitter);
exports.NoSqlDataProvider = NoSqlDataProvider;
