/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2013-2014 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of the GNU General
 * Public License Version 2 only ("GPL"). You may not use this file except
 * in compliance with the License.  You can obtain a copy of the License at
 * https://avatar.java.net/license.html or legal/LICENSE.txt.  See the
 * License for the specific language governing permissions and limitations
 * under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 */

/**
 * This module provides a node based integration with the platform's JMS system. It
 * is supported only in Avatar EE on Glassfish</b>.

 * @module org/glassfish/avatar/jms
 */

'use strict';

var events = require("events");
var util = require("util");
var common = require('org/glassfish/avatar/common');
var Promise = common.Promise;
var wrapCallback = common.wrapCallback;
var helperFactory = Java.type('org.glassfish.avatar.runtime.service.jms.JMSHelperFactory').newFactory(serviceContainer);

/**
 * Creates a JMSFactory. 
 * @param {object} properties An objet with properties to the JMSFactory. Currently the only
 * property supported is "name", which corresponds to the name of the JMSConnectionFactory 
 * configured in the system (typically a JNDI name). If this parameter is a string it will be used as the name property.
 * @constructor
 * 
 * @classdesc A JMSFactory provides access to a JMS provider and lets you create JMS message 
 * producers and consumers to communicate with JMS queues and topics. The JMSFactory encapsulates the ConnectionFactory, 
 * Connection, and Session interfaces in JMS.
 * @example 
 * var myJMSFactory = new JMSFactory("jms/MyConnectionFactory");
 * @example
 * var myJMSFactory = new JMSFactory({name: "jms/MyConnectionFactory"});
 * @private
 */
var JMSFactory = function(name){
    var helper = helperFactory.getJMSHelper(name);

    /**
     * Callback for asynchronous factory methods
     * @callback JMSFactory~callback
     * @param {string} error If an error occurred, otherwise null
     * @param {object} result The object that was created
     */
    
    /**
     * Creates a producer for sending messages on the given destination.
     * 
     * @param {string} destinationName the JMS topic or queue name to send messages to.
     * @param {JMSFactory~callback} callback (optional)
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.createProducer = function(destinationName, callback){
        var deferred = Promise.defer();
        helper.createProducer(destinationName, function(event, args){
            deferred.resolve(new JMSProducer(helper, args[0], args[1]));//args=[producer, session]
        }, function(event, args){
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };
    
    /**
     * Creates a producer for receiving messages on the given destination.
     * 
     * @param {string} destinationName the JMS topic or queue name to send messages to.
     * @param {string} messageSelector a JMS messageSelector used to filter messages.
     * @param {JMSFactory~callback} callback (optional)
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.createConsumer = function(destinationName, messageSelector, callback){
        var deferred = Promise.defer();
        helper.createConsumer(destinationName, messageSelector || null, function(event, args){
            deferred.resolve(new JMSConsumer(helper, args[0], args[1]));//args=[consumer, connection]
        }, function(event, args){
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    }
};
    

 /**
 * @classdesc A JMSProducer is used to send messages to a specific destination. 
 * JMSProducer objects are created with the {@link JMSFactory}.
 * A message can have properties that can be used by consumers to select specific messages
 * using a messageSelector.
 * 
 * @example <caption>This example illustrates how to create a producer that sends a hello world 
 * message to a topic.</caption>
 * //Create the JMSProducer for destination "myTopic"
 * myFactory.createProducer("myTopic")
 * .then(function(producer){
 *   //Send a message using the created producer.
 *   return producer.send("Hello World!");
 * });
 * 
 * @constructor
 * @private
 */
var JMSProducer = function(helper, producer, session){
    
    /**
     * Sends a message to the JMS destination.
     * 
     * @param {string} message the message to send
     * @param {object} properties a properties object that will be associated with every message
     * and can be used for dispatching together with message selectors. null means the message will
     * have no properties.
     * @param {noargCallback} callback (optional) 
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.send = function(message, properties, callback){
        var deferred = Promise.defer();
        helper.send(producer, session, message, properties || null, function(){
            deferred.resolve();
        }, function(event, args){
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };
    
    /**
     * Closes this JMSProducer, releasing all it's resources. 
     * A closed JMSProducer cannot be used to send messages.
     * @param {noargCallback} callback (optional)
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.close = function(callback){//TODO: The connection has to be closed too!
        var deferred = Promise.defer();
        helper.close(session, function(){
            deferred.resolve();
        }, function(event, args){
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };
};

/**
 * This constructor is for internal use only.
 * 
 * @classdesc A JMSConsumer is used to subscribe for and receive messages from a specific destination.
 * JMSConsumer objects are created with the {@link JMSFactory}. A consumer is created with an optional
 * messageSelector to receive only messages with specific properties.
 * This object implements the EventEmitter API.
 * 
 * @example <caption>This example illustrates how to create a consumer that logs any 
 * incoming message from the queue. </caption>
 * //Create a consumer on destination "myQueue" (The destination has to be configured in the JMS system)
 * factory.createConsumer("myQueue")
 * .then(function(consumer){
 *   //Register a listener
 *   consumer.on("message", function(message){
 *     avatar.log("I got a message! "+message)
 *   });
 * });
 * 
 * @class
 * @private
 */
var JMSConsumer = function(helper, consumer, connection){
    
    /**
     * Closes this JMS Consumer, releasing all it's system resources.
     * A closed JMSConsumer cannot receive any new messages.
     * @param {noargCallback} callback (optional)
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.close = function(callback){
        this.removeAllListeners("message");
        var deferred = Promise.defer();
        helper.close(connection, function(){
            deferred.resolve();
        }, function(event, args){
            deferred.reject(args[0]);
        });
        return wrapCallback(callback, deferred.promise);
    };

    // EventEmitter methods
    
    var that = this;
    var superAddListener = this.addListener;
    /**
     * Adds a listener to this JMSConsumer.
     * @param {noargCallback} callback (optional)
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.addListener = function(event, listener, callback){
        var deferred = Promise.defer();
        if(event === "message" && this.listeners(event).length === 0){
            helper.setListener(connection, consumer, function(event, args){
                /**
                 * When a message is available for this consumer of the destination.
                 * @event JMSConsumer#message
                 * @type {string}
                 * @private
                 */
                that.emit("message", args[0]);
            }, function(){
                deferred.resolve();
            }, function(event, args){
                deferred.reject(args[0]);
            });
        } else {
            process.nextTick(function(){
                deferred.resolve();
            });
        }
        superAddListener.call(this, event, listener);
        return wrapCallback(callback, deferred.promise);
    };
    
    this.on = this.addListener;
    
    var superRemoveListener = this.removeListener;
    /**
     * Removes a listener from this JMSConsumer.
     * @param {noargCallback} callback (optional)
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.removeListener = function(event, listener, callback) {
        var deferred = Promise.defer();
        superRemoveListener.call(this, event, listener);
        if(event === "message" && this.listeners(event).length === 0){
            helper.setListener(connection, consumer, null, function(){
                deferred.resolve();
            }, function(event, args){
                deferred.reject(args[0]);
            });
        } else {
            process.nextTick(function(){
                deferred.resolve();
            });
        }
        return wrapCallback(callback, deferred.promise);
    };
};

util.inherits(JMSConsumer, events.EventEmitter);        

/**
 * Callback for asynchronous methods indicating that the operation has completed or a failure has occurred.
 * @callback noargCallback
 * @param {string} error If an error occurred, otherwise null
 */

/**
 * @class
 * @classdesc A JavaScript interface for sending and receiving messages on a JMS system. It acts as a JMS consumer and producer
 * and supports message properties and messageSelectors. 
 * <p><b>Note that <tt>JMS</tt> is supported only in Avatar GF</b>.
 *
 * @param {object} properties An object containing the following properties:
 *  destinationName - the name of the JMSDestination (typically a JNDI name configured in the system),
 *  connectionFactoryName - the name of the JMSConnectionFactory (typically a JNDI name configured in the system)
 * @static
 *
 *  @example <caption>This example illustrates how to add a listener and send a message to a JMS destination.</caption>
 *  var myJMS = new avatar.JMS({connectionFactoryName: "MyConnectionFactory", destinationName: "MyTopic"});
 *  avatar.log("Adding listener...");
 *  myJMS.addListener(function(message){
 *      avatar.log("Got message: "+message);
 *  })
 *  .then(function(){
 *      avatar.log("Sending message...");
 *      return myJMS.send("Test message.");
 *  })
 *  .then(function(){
 *      avatar.log("Message sent.");
 *  });
 */
var JMS = function(properties){
    this.ELCompilerIgnore = true;
    
    var factory = new JMSFactory(properties.connectionFactoryName);
    var producer;
    var consumer;
 
    /**
     * Sends a message to the JMS destination.
     * 
     * @param {string} message the message to send
     * @param {object} properties a properties object that will be associated with every message
     * and can be used for dispatching together with message selectors. null means the message will
     * have no properties.
     * @param {noargCallback} callback (optional) 
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.send = function(message, props, callback){
        return wrapCallback(callback,
            Promise(producer || factory.createProducer(properties.destinationName)).then(function(p){
                producer = p;
                return producer.send(message, props);
            })
        );
    };
    
    /**
     * Adds a listener to this JMS instance.
     * @param {JMS~listener} listener the message listener to register 
     * @param {noargCallback} callback (optional)
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.addListener = function(listener, callback){
        return wrapCallback(callback,
            Promise(consumer || factory.createConsumer(properties.destinationName, properties.messageSelector)).then(function(c){
                consumer = c;
                return consumer.on("message", listener);
            })
        );
    };
    
    /**
     * Removes a listener from this JMS instance.
     * @param {JMS~listener} listener to remove
     * @param {noargCallback} callback (optional)
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.removeListener = function(listener, callback){
        return wrapCallback(callback,
                consumer.removeListener("message", listener));
    };
    
    /**
     * Closes all resources associated with this JMS instance.
     * @returns a promise unless a callback is provided. See {@link http://documentup.com/kriskowal/q/}.
     */
    this.close = function(callback){
        var promises = [];
        if(producer) {
            promises.push(producer.close());
        }
        if(consumer) {
            promises.push(consumer.close());
        }
        return wrapCallback(callback,Promise.all(promises));
    };

    /**
     * Listens for JMS messages from a JMS queue or topic.
     * @callback JMS~listener
     * @param message the received message
     */    
};

exports.JMS = JMS;
